@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2024 The pgagroal community
+ * Copyright (C) 2026 The pgagroal community
  *
  * Redistribution and use in source and binary forms, with or without modification,
  * are permitted provided that the following conditions are met:
@@ -28,11 +28,14 @@
 
 /* pgagroal */
 #include <pgagroal.h>
+#include <ev.h>
 #include <logging.h>
 #include <memory.h>
 #include <message.h>
 #include <network.h>
+#include <shmem.h>
 #include <utils.h>
+#include <worker.h>
 
 #include <assert.h>
 #include <errno.h>
@@ -93,6 +96,107 @@ pgagroal_write_socket_message(int socket, struct message* msg)
    return write_message(socket, msg);
 }
 
+static int
+read_message_from_buffer(struct io_watcher* watcher, struct message** msg_p)
+{
+   struct message* msg = pgagroal_get_watcher_message(watcher);
+
+   if (msg->length == 0)
+   {
+      return MESSAGE_STATUS_ZERO;
+   }
+
+   if (msg->length < 0)
+   {
+      return MESSAGE_STATUS_ERROR;
+   }
+
+   msg->kind = (signed char)(*((char*)msg->data));
+   *msg_p = msg;
+   return MESSAGE_STATUS_OK;
+}
+
+static int
+write_message_from_buffer(struct io_watcher* watcher, struct message* msg)
+{
+   int sent_bytes = pgagroal_event_prep_submit_send(watcher, msg);
+
+   if (msg->length == 0)
+   {
+      return MESSAGE_STATUS_ZERO;
+   }
+   if (sent_bytes < msg->length)
+   {
+      return MESSAGE_STATUS_ERROR;
+   }
+   return MESSAGE_STATUS_OK;
+}
+
+static int __attribute__((unused))
+read_from_buffer(struct io_watcher* watcher __attribute__((unused)), struct message* msg)
+{
+   int read_bytes = pgagroal_wait_recv();
+
+   ((struct message*)msg)->length = read_bytes;
+   msg->kind = (signed char)(*((char*)msg->data));
+
+   if (msg->length == 0)
+   {
+      return MESSAGE_STATUS_ZERO;
+   }
+   return MESSAGE_STATUS_OK;
+}
+
+int
+pgagroal_recv_message(struct io_watcher* watcher, struct message** msg)
+{
+   struct main_configuration* config = (struct main_configuration*)shmem;
+   struct worker_io* wi = (struct worker_io*)watcher;
+
+   int rfd = watcher->fds.worker.rcv_fd;
+
+   /* Use the correct TLS context for the receiving endpoint */
+   if (rfd == wi->client_fd && wi->client_ssl != NULL)
+   {
+      return ssl_read_message(wi->client_ssl, 0, msg);
+   }
+   if (rfd == wi->server_fd && wi->server_ssl != NULL)
+   {
+      return ssl_read_message(wi->server_ssl, 0, msg);
+   }
+
+   if (config->ev_backend != PGAGROAL_EVENT_BACKEND_IO_URING)
+   {
+      return read_message(rfd, false, 0, msg);
+   }
+   return read_message_from_buffer(watcher, msg);
+}
+
+int
+pgagroal_send_message(struct io_watcher* watcher, struct message* msg)
+{
+   struct main_configuration* config = (struct main_configuration*)shmem;
+   struct worker_io* wi = (struct worker_io*)watcher;
+
+   int sfd = watcher->fds.worker.snd_fd;
+
+   /* Use the correct TLS context for the sending endpoint */
+   if (sfd == wi->client_fd && wi->client_ssl != NULL)
+   {
+      return ssl_write_message(wi->client_ssl, msg);
+   }
+   if (sfd == wi->server_fd && wi->server_ssl != NULL)
+   {
+      return ssl_write_message(wi->server_ssl, msg);
+   }
+
+   if (config->ev_backend != PGAGROAL_EVENT_BACKEND_IO_URING)
+   {
+      return write_message(sfd, msg);
+   }
+   return write_message_from_buffer(watcher, msg);
+}
+
 int
 pgagroal_read_ssl_message(SSL* ssl, struct message** msg)
 {
@@ -134,7 +238,7 @@ pgagroal_create_message(void* data, ssize_t length, struct message** msg)
 }
 
 void
-pgagroal_free_message(struct message* msg)
+pgagroal_clear_message(struct message* msg __attribute__((unused)))
 {
    pgagroal_memory_free();
 }
@@ -173,7 +277,7 @@ pgagroal_copy_message(struct message* msg)
 }
 
 void
-pgagroal_free_copy_message(struct message* msg)
+pgagroal_free_message(struct message* msg)
 {
    if (msg)
    {
@@ -512,14 +616,14 @@ pgagroal_write_deallocate_all(SSL* ssl, int socket)
       goto error;
    }
 
-   pgagroal_free_message(reply);
+   pgagroal_clear_message(reply);
 
    return 0;
 
 error:
    if (reply)
    {
-      pgagroal_free_message(reply);
+      pgagroal_clear_message(reply);
    }
 
    return 1;
@@ -559,6 +663,10 @@ pgagroal_write_discard_all(SSL* ssl, int socket)
       goto error;
    }
 
+   /* XXX: someone is destroying the memory before reaching here in the worker.
+    * Allocate another buffer for now, but this needs fixing. */
+   pgagroal_memory_init();
+
    if (ssl == NULL)
    {
       status = read_message(socket, true, 0, &reply);
@@ -577,14 +685,14 @@ pgagroal_write_discard_all(SSL* ssl, int socket)
       goto error;
    }
 
-   pgagroal_free_message(reply);
+   pgagroal_clear_message(reply);
 
    return 0;
 
 error:
    if (reply)
    {
-      pgagroal_free_message(reply);
+      pgagroal_clear_message(reply);
    }
 
    return 1;
@@ -715,14 +823,14 @@ pgagroal_write_rollback(SSL* ssl, int socket)
    {
       goto error;
    }
-   pgagroal_free_message(reply);
+   pgagroal_clear_message(reply);
 
    return 0;
 
 error:
    if (reply)
    {
-      pgagroal_free_message(reply);
+      pgagroal_clear_message(reply);
    }
 
    return 1;
@@ -1174,15 +1282,20 @@ pgagroal_connection_isvalid(int socket)
    {
       goto error;
    }
+   if (reply->kind == 'e')
+   {
+      pgagroal_log_error("Unexpected message kind");
+      goto error;
+   }
 
-   pgagroal_free_message(reply);
+   pgagroal_clear_message(reply);
 
    return true;
 
 error:
    if (reply)
    {
-      pgagroal_free_message(reply);
+      pgagroal_clear_message(reply);
    }
 
    return false;
@@ -1212,6 +1325,7 @@ read_message(int socket, bool block, int timeout, struct message** msg)
    ssize_t numbytes;
    struct timeval tv;
    struct message* m = NULL;
+   pgagroal_memory_init();
 
    if (unlikely(timeout > 0))
    {
@@ -1223,10 +1337,8 @@ read_message(int socket, bool block, int timeout, struct message** msg)
    do
    {
       keep_read = false;
-
       m = pgagroal_memory_message();
-
-      numbytes = read(socket, m->data, m->max_length);
+      numbytes = read(socket, m->data, MESSAGE_PARSE_BUFFER_SIZE);
 
       if (likely(numbytes > 0))
       {
@@ -1258,7 +1370,6 @@ read_message(int socket, bool block, int timeout, struct message** msg)
                tv.tv_usec = 0;
                setsockopt(socket, SOL_SOCKET, SO_RCVTIMEO, (const char*)&tv, sizeof(tv));
             }
-
             return MESSAGE_STATUS_ZERO;
          }
       }
@@ -1269,6 +1380,10 @@ read_message(int socket, bool block, int timeout, struct message** msg)
             keep_read = true;
             errno = 0;
          }
+         else
+         {
+            pgagroal_log_error("read error: fd=%d errno=%d", socket, errno);
+         }
       }
    }
    while (keep_read);
@@ -1353,6 +1468,8 @@ ssl_read_message(SSL* ssl, int timeout, struct message** msg)
    time_t start_time;
    struct message* m = NULL;
 
+   pgagroal_memory_init();
+
    if (unlikely(timeout > 0))
    {
       start_time = time(NULL);
@@ -1364,7 +1481,7 @@ ssl_read_message(SSL* ssl, int timeout, struct message** msg)
 
       m = pgagroal_memory_message();
 
-      numbytes = SSL_read(ssl, m->data, m->max_length);
+      numbytes = SSL_read(ssl, m->data, MESSAGE_PARSE_BUFFER_SIZE);
 
       if (likely(numbytes > 0))
       {
@@ -1386,7 +1503,6 @@ ssl_read_message(SSL* ssl, int timeout, struct message** msg)
             case SSL_ERROR_ZERO_RETURN:
                if (timeout > 0)
                {
-
                   if (difftime(time(NULL), start_time) >= timeout)
                   {
                      return MESSAGE_STATUS_ZERO;
@@ -1394,21 +1510,17 @@ ssl_read_message(SSL* ssl, int timeout, struct message** msg)
 
                   /* Sleep for 100ms */
                   SLEEP(100000000L)
-
                }
+               __attribute__((fallthrough));
             case SSL_ERROR_WANT_READ:
             case SSL_ERROR_WANT_WRITE:
             case SSL_ERROR_WANT_CONNECT:
             case SSL_ERROR_WANT_ACCEPT:
             case SSL_ERROR_WANT_X509_LOOKUP:
 #ifndef HAVE_OPENBSD
-#if (OPENSSL_VERSION_NUMBER >= 0x10100000L)
             case SSL_ERROR_WANT_ASYNC:
             case SSL_ERROR_WANT_ASYNC_JOB:
-#if (OPENSSL_VERSION_NUMBER >= 0x10101000L)
             case SSL_ERROR_WANT_CLIENT_HELLO_CB:
-#endif
-#endif
 #endif
                keep_read = true;
                break;
@@ -1486,13 +1598,9 @@ ssl_write_message(SSL* ssl, struct message* msg)
             case SSL_ERROR_WANT_ACCEPT:
             case SSL_ERROR_WANT_X509_LOOKUP:
 #ifndef HAVE_OPENBSD
-#if (OPENSSL_VERSION_NUMBER >= 0x10100000L)
             case SSL_ERROR_WANT_ASYNC:
             case SSL_ERROR_WANT_ASYNC_JOB:
-#if (OPENSSL_VERSION_NUMBER >= 0x10101000L)
             case SSL_ERROR_WANT_CLIENT_HELLO_CB:
-#endif
-#endif
 #endif
                errno = 0;
                keep_write = true;
@@ -1524,3 +1632,14 @@ ssl_write_message(SSL* ssl, struct message* msg)
 
    return MESSAGE_STATUS_ERROR;
 }
+
+struct message*
+pgagroal_get_watcher_message(struct io_watcher* watcher)
+{
+   if (watcher != NULL && watcher->msg != NULL)
+   {
+      return watcher->msg;
+   }
+
+   return pgagroal_memory_message();
+}
