@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2024 The pgagroal community
+ * Copyright (C) 2026 The pgagroal community
  *
  * Redistribution and use in source and binary forms, with or without modification,
  * are permitted provided that the following conditions are met:
@@ -28,6 +28,7 @@
 
 /* pgagroal */
 #include <pgagroal.h>
+#include <ev.h>
 #include <logging.h>
 #include <management.h>
 #include <message.h>
@@ -41,7 +42,7 @@
 
 /* system */
 #include <errno.h>
-#include <ev.h>
+#include <signal.h>
 #include <stdatomic.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -49,11 +50,11 @@
 #include <sys/types.h>
 #include <sys/socket.h>
 
-static int  session_initialize(void*, void**, size_t*);
-static void session_start(struct ev_loop* loop, struct worker_io*);
-static void session_client(struct ev_loop* loop, struct ev_io* watcher, int revents);
-static void session_server(struct ev_loop* loop, struct ev_io* watcher, int revents);
-static void session_stop(struct ev_loop* loop, struct worker_io*);
+static int session_initialize(void*, void**, size_t*);
+static void session_start(struct event_loop* loop, struct worker_io*);
+static void session_client(struct io_watcher* watcher);
+static void session_server(struct io_watcher* watcher);
+static void session_stop(struct event_loop* loop, struct worker_io*);
 static void session_destroy(void*, size_t);
 static void session_periodic(void);
 
@@ -98,9 +99,9 @@ session_initialize(void* shmem, void** pipeline_shmem, size_t* pipeline_shmem_si
    void* session_shmem = NULL;
    size_t session_shmem_size;
    struct client_session* client;
-   struct configuration* config;
+   struct main_configuration* config;
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    *pipeline_shmem = NULL;
    *pipeline_shmem_size = 0;
@@ -108,7 +109,7 @@ session_initialize(void* shmem, void** pipeline_shmem, size_t* pipeline_shmem_si
    if (config->disconnect_client > 0)
    {
       session_shmem_size = config->max_connections * sizeof(struct client_session);
-      if (pgagroal_create_shared_memory(session_shmem_size, config->hugepage, &session_shmem))
+      if (pgagroal_create_shared_memory(session_shmem_size, config->common.hugepage, &session_shmem))
       {
          return 1;
       }
@@ -130,12 +131,12 @@ session_initialize(void* shmem, void** pipeline_shmem, size_t* pipeline_shmem_si
 }
 
 static void
-session_start(struct ev_loop* loop, struct worker_io* w)
+session_start(struct event_loop* loop __attribute__((unused)), struct worker_io* w)
 {
    struct client_session* client;
-   struct configuration* config;
+   struct main_configuration* config;
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    in_tx = false;
    next_client_message = 0;
@@ -161,7 +162,7 @@ session_start(struct ev_loop* loop, struct worker_io* w)
 }
 
 static void
-session_stop(struct ev_loop* loop, struct worker_io* w)
+session_stop(struct event_loop* loop __attribute__((unused)), struct worker_io* w)
 {
    struct client_session* client;
 
@@ -195,9 +196,9 @@ session_periodic(void)
    int socket;
    struct message* cancel_msg = NULL;
    struct client_session* client;
-   struct configuration* config;
+   struct main_configuration* config;
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    if (config->disconnect_client > 0 && pipeline_shmem != NULL)
    {
@@ -244,7 +245,7 @@ session_periodic(void)
                   }
                   else
                   {
-                     ret = pgagroal_connect(config->servers[server].host, config->servers[server].port, &socket);
+                     ret = pgagroal_connect(config->servers[server].host, config->servers[server].port, &socket, config->keep_alive, config->nodelay);
                   }
 
                   if (ret == 0)
@@ -265,7 +266,7 @@ session_periodic(void)
                                     i, config->connections[i].pid, config->connections[i].fd);
                   kill(config->connections[i].pid, SIGQUIT);
 
-                  pgagroal_free_copy_message(cancel_msg);
+                  pgagroal_free_message(cancel_msg);
                   cancel_msg = NULL;
                }
                else
@@ -281,26 +282,20 @@ session_periodic(void)
 }
 
 static void
-session_client(struct ev_loop* loop, struct ev_io* watcher, int revents)
+session_client(struct io_watcher* watcher)
 {
    int status = MESSAGE_STATUS_ERROR;
    struct worker_io* wi = NULL;
    struct message* msg = NULL;
-   struct configuration* config = NULL;
+   struct main_configuration* config = NULL;
 
    wi = (struct worker_io*)watcher;
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    client_active(wi->slot);
 
-   if (wi->client_ssl == NULL)
-   {
-      status = pgagroal_read_socket_message(wi->client_fd, &msg);
-   }
-   else
-   {
-      status = pgagroal_read_ssl_message(wi->client_ssl, &msg);
-   }
+   status = pgagroal_recv_message(watcher, &msg);
+
    if (likely(status == MESSAGE_STATUS_OK))
    {
       pgagroal_prometheus_network_sent_add(msg->length);
@@ -342,14 +337,8 @@ session_client(struct ev_loop* loop, struct ev_io* watcher, int revents)
             }
          }
 
-         if (wi->server_ssl == NULL)
-         {
-            status = pgagroal_write_socket_message(wi->server_fd, msg);
-         }
-         else
-         {
-            status = pgagroal_write_ssl_message(wi->server_ssl, msg);
-         }
+         status = pgagroal_send_message(watcher, msg);
+
          if (unlikely(status == MESSAGE_STATUS_ERROR))
          {
             if (config->failover)
@@ -369,7 +358,7 @@ session_client(struct ev_loop* loop, struct ev_io* watcher, int revents)
       else if (msg->kind == 'X')
       {
          saw_x = true;
-         running = 0;
+         pgagroal_event_loop_break();
       }
    }
    else if (status == MESSAGE_STATUS_ZERO)
@@ -383,7 +372,6 @@ session_client(struct ev_loop* loop, struct ev_io* watcher, int revents)
 
    client_inactive(wi->slot);
 
-   ev_break(loop, EVBREAK_ONE);
    return;
 
 client_done:
@@ -403,8 +391,7 @@ client_done:
       exit_code = WORKER_SERVER_FAILURE;
    }
 
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+   pgagroal_event_loop_break();
    return;
 
 client_error:
@@ -417,8 +404,8 @@ client_error:
    client_inactive(wi->slot);
 
    exit_code = WORKER_CLIENT_FAILURE;
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+
+   pgagroal_event_loop_break();
    return;
 
 server_error:
@@ -431,8 +418,8 @@ server_error:
    client_inactive(wi->slot);
 
    exit_code = WORKER_SERVER_FAILURE;
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+
+   pgagroal_event_loop_break();
    return;
 
 failover:
@@ -440,32 +427,26 @@ failover:
    client_inactive(wi->slot);
 
    exit_code = WORKER_FAILOVER;
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+
+   pgagroal_event_loop_break();
    return;
 }
 
 static void
-session_server(struct ev_loop* loop, struct ev_io* watcher, int revents)
+session_server(struct io_watcher* watcher)
 {
    int status = MESSAGE_STATUS_ERROR;
    bool fatal = false;
    struct worker_io* wi = NULL;
    struct message* msg = NULL;
-   struct configuration* config = NULL;
+   struct main_configuration* config = (struct main_configuration*)shmem;
 
    wi = (struct worker_io*)watcher;
 
    client_active(wi->slot);
 
-   if (wi->server_ssl == NULL)
-   {
-      status = pgagroal_read_socket_message(wi->server_fd, &msg);
-   }
-   else
-   {
-      status = pgagroal_read_ssl_message(wi->server_ssl, &msg);
-   }
+   status = pgagroal_recv_message(watcher, &msg);
+
    if (likely(status == MESSAGE_STATUS_OK))
    {
       pgagroal_prometheus_network_received_add(msg->length);
@@ -510,14 +491,9 @@ session_server(struct ev_loop* loop, struct ev_io* watcher, int revents)
             next_server_message -= offset;
          }
       }
-      if (wi->client_ssl == NULL)
-      {
-         status = pgagroal_write_socket_message(wi->client_fd, msg);
-      }
-      else
-      {
-         status = pgagroal_write_ssl_message(wi->client_ssl, msg);
-      }
+
+      status = pgagroal_send_message(watcher, msg);
+
       if (unlikely(status != MESSAGE_STATUS_OK))
       {
          goto client_error;
@@ -535,7 +511,7 @@ session_server(struct ev_loop* loop, struct ev_io* watcher, int revents)
          if (fatal)
          {
             exit_code = WORKER_SERVER_FATAL;
-            running = 0;
+            pgagroal_event_loop_break();
          }
       }
    }
@@ -550,11 +526,9 @@ session_server(struct ev_loop* loop, struct ev_io* watcher, int revents)
 
    client_inactive(wi->slot);
 
-   ev_break(loop, EVBREAK_ONE);
    return;
 
 client_error:
-   config = (struct configuration*)shmem;
    pgagroal_log_warn("[S] Client error (slot %d database %s user %s): %s (socket %d status %d)",
                      wi->slot, config->connections[wi->slot].database, config->connections[wi->slot].username,
                      strerror(errno), wi->client_fd, status);
@@ -564,12 +538,11 @@ client_error:
    client_inactive(wi->slot);
 
    exit_code = WORKER_CLIENT_FAILURE;
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+
+   pgagroal_event_loop_break();
    return;
 
 server_done:
-   config = (struct configuration*)shmem;
    pgagroal_log_debug("[S] Server done (slot %d database %s user %s): %s (socket %d status %d)",
                       wi->slot, config->connections[wi->slot].database, config->connections[wi->slot].username,
                       strerror(errno), wi->server_fd, status);
@@ -577,12 +550,10 @@ server_done:
 
    client_inactive(wi->slot);
 
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+   pgagroal_event_loop_break();
    return;
 
 server_error:
-   config = (struct configuration*)shmem;
    pgagroal_log_warn("[S] Server error (slot %d database %s user %s): %s (socket %d status %d)",
                      wi->slot, config->connections[wi->slot].database, config->connections[wi->slot].username,
                      strerror(errno), wi->server_fd, status);
@@ -592,8 +563,8 @@ server_error:
    client_inactive(wi->slot);
 
    exit_code = WORKER_SERVER_FAILURE;
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+
+   pgagroal_event_loop_break();
    return;
 }
 
