@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2024 The pgagroal community
+ * Copyright (C) 2026 The pgagroal community
  *
  * Redistribution and use in source and binary forms, with or without modification,
  * are permitted provided that the following conditions are met:
@@ -28,8 +28,9 @@
 
 /* pgagroal */
 #include <pgagroal.h>
+#include <connection.h>
+#include <ev.h>
 #include <logging.h>
-#include <management.h>
 #include <message.h>
 #include <network.h>
 #include <pipeline.h>
@@ -43,24 +44,23 @@
 
 /* system */
 #include <errno.h>
-#include <ev.h>
 #include <stdlib.h>
 #include <unistd.h>
 #include <netinet/in.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 
-static int  transaction_initialize(void*, void**, size_t*);
-static void transaction_start(struct ev_loop* loop, struct worker_io*);
-static void transaction_client(struct ev_loop* loop, struct ev_io* watcher, int revents);
-static void transaction_server(struct ev_loop* loop, struct ev_io* watcher, int revents);
-static void transaction_stop(struct ev_loop* loop, struct worker_io*);
+static int transaction_initialize(void*, void**, size_t*);
+static void transaction_start(struct event_loop* loop, struct worker_io*);
+static void transaction_client(struct io_watcher* watcher);
+static void transaction_server(struct io_watcher* watcher);
+static void transaction_stop(struct event_loop* loop, struct worker_io*);
 static void transaction_destroy(void*, size_t);
 static void transaction_periodic(void);
 
-static void start_mgt(struct ev_loop* loop);
-static void shutdown_mgt(struct ev_loop* loop);
-static void accept_cb(struct ev_loop* loop, struct ev_io* watcher, int revents);
+static void start_mgt(struct event_loop* loop);
+static void shutdown_mgt(struct event_loop* loop);
+static void accept_cb(struct io_watcher* watcher);
 
 static int slot;
 static char username[MAX_USERNAME_LENGTH];
@@ -74,8 +74,9 @@ static int deallocate;
 static bool fatal;
 static int fds[MAX_NUMBER_OF_CONNECTIONS];
 static bool saw_x = false;
-static struct ev_io io_mgt;
+static struct io_watcher io_mgt;
 static struct worker_io server_io;
+static bool io_watcher_active = false;
 
 struct pipeline
 transaction_pipeline(void)
@@ -94,19 +95,19 @@ transaction_pipeline(void)
 }
 
 static int
-transaction_initialize(void* shmem, void** pipeline_shmem, size_t* pipeline_shmem_size)
+transaction_initialize(void* shmem __attribute__((unused)), void** pipeline_shmem __attribute__((unused)), size_t* pipeline_shmem_size __attribute__((unused)))
 {
    return 0;
 }
 
 static void
-transaction_start(struct ev_loop* loop, struct worker_io* w)
+transaction_start(struct event_loop* loop, struct worker_io* w)
 {
    char p[MISC_LENGTH];
    bool is_new;
-   struct configuration* config = NULL;
+   struct main_configuration* config = NULL;
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    slot = -1;
    memcpy(&username[0], config->connections[w->slot].username, MAX_USERNAME_LENGTH);
@@ -118,7 +119,7 @@ transaction_start(struct ev_loop* loop, struct worker_io* w)
    deallocate = false;
 
    memset(&p, 0, sizeof(p));
-   snprintf(&p[0], sizeof(p), ".s.%d", getpid());
+   snprintf(&p[0], sizeof(p), ".s.pgagroal.%d", getpid());
 
    if (pgagroal_bind_unix_socket(config->unix_socket_dir, &p[0], &unix_socket))
    {
@@ -152,28 +153,31 @@ transaction_start(struct ev_loop* loop, struct worker_io* w)
 error:
 
    exit_code = WORKER_FAILURE;
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+   pgagroal_event_loop_break();
    return;
 }
 
 static void
-transaction_stop(struct ev_loop* loop, struct worker_io* w)
+transaction_stop(struct event_loop* loop, struct worker_io* w)
 {
    if (slot != -1)
    {
-      struct configuration* config = NULL;
+      struct main_configuration* config = NULL;
 
-      config = (struct configuration*)shmem;
+      config = (struct main_configuration*)shmem;
 
       /* We are either in 'X' or the client terminated (consider cancel query) */
       if (in_tx)
       {
          /* ROLLBACK */
-         pgagroal_write_rollback(NULL, config->connections[slot].fd);
+         pgagroal_write_rollback(w->server_ssl, config->connections[slot].fd);
       }
 
-      ev_io_stop(loop, (struct ev_io*)&server_io);
+      if (io_watcher_active)
+      {
+         pgagroal_io_stop(&server_io.io);
+         io_watcher_active = false;
+      }
       pgagroal_tracking_event_slot(TRACKER_TX_RETURN_CONNECTION_STOP, w->slot);
       pgagroal_return_connection(slot, w->server_ssl, true);
       slot = -1;
@@ -183,7 +187,7 @@ transaction_stop(struct ev_loop* loop, struct worker_io* w)
 }
 
 static void
-transaction_destroy(void* pipeline_shmem, size_t pipeline_shmem_size)
+transaction_destroy(void* pipeline_shmem __attribute__((unused)), size_t pipeline_shmem_size __attribute__((unused)))
 {
 }
 
@@ -193,16 +197,16 @@ transaction_periodic(void)
 }
 
 static void
-transaction_client(struct ev_loop* loop, struct ev_io* watcher, int revents)
+transaction_client(struct io_watcher* watcher)
 {
    int status = MESSAGE_STATUS_ERROR;
    SSL* s_ssl = NULL;
    struct worker_io* wi = NULL;
    struct message* msg = NULL;
-   struct configuration* config = NULL;
+   struct main_configuration* config = NULL;
 
    wi = (struct worker_io*)watcher;
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    /* We can't use the information from wi except from client_fd/client_ssl */
    if (slot == -1)
@@ -214,13 +218,16 @@ transaction_client(struct ev_loop* loop, struct ev_io* watcher, int revents)
          goto get_error;
       }
 
-      wi->server_fd = config->connections[slot].fd;
+      wi->server_fd = fds[slot];
       wi->server_ssl = s_ssl;
       wi->slot = slot;
 
+      pgagroal_event_worker_init(&wi->io, wi->client_fd, wi->server_fd, transaction_client);
+
       memcpy(&config->connections[slot].appname[0], &appname[0], MAX_APPLICATION_NAME);
 
-      ev_io_init((struct ev_io*)&server_io, transaction_server, config->connections[slot].fd, EV_READ);
+      pgagroal_event_worker_init(&server_io.io, config->connections[slot].fd,
+                                 wi->client_fd, transaction_server);
       server_io.client_fd = wi->client_fd;
       server_io.server_fd = config->connections[slot].fd;
       server_io.slot = slot;
@@ -229,17 +236,12 @@ transaction_client(struct ev_loop* loop, struct ev_io* watcher, int revents)
 
       fatal = false;
 
-      ev_io_start(loop, (struct ev_io*)&server_io);
+      pgagroal_io_start(&server_io.io);
+      io_watcher_active = true;
    }
 
-   if (wi->client_ssl == NULL)
-   {
-      status = pgagroal_read_socket_message(wi->client_fd, &msg);
-   }
-   else
-   {
-      status = pgagroal_read_ssl_message(wi->client_ssl, &msg);
-   }
+   status = pgagroal_recv_message(watcher, &msg);
+
    if (likely(status == MESSAGE_STATUS_OK))
    {
       pgagroal_prometheus_network_sent_add(msg->length);
@@ -294,14 +296,8 @@ transaction_client(struct ev_loop* loop, struct ev_io* watcher, int revents)
             }
          }
 
-         if (wi->server_ssl == NULL)
-         {
-            status = pgagroal_write_socket_message(wi->server_fd, msg);
-         }
-         else
-         {
-            status = pgagroal_write_ssl_message(wi->server_ssl, msg);
-         }
+         status = pgagroal_send_message(watcher, msg);
+
          if (unlikely(status == MESSAGE_STATUS_ERROR))
          {
             if (config->failover)
@@ -321,7 +317,7 @@ transaction_client(struct ev_loop* loop, struct ev_io* watcher, int revents)
       else if (msg->kind == 'X')
       {
          saw_x = true;
-         running = 0;
+         pgagroal_event_loop_break();
       }
    }
    else if (status == MESSAGE_STATUS_ZERO)
@@ -333,7 +329,6 @@ transaction_client(struct ev_loop* loop, struct ev_io* watcher, int revents)
       goto client_error;
    }
 
-   ev_break(loop, EVBREAK_ONE);
    return;
 
 client_done:
@@ -351,8 +346,7 @@ client_done:
       exit_code = WORKER_SERVER_FAILURE;
    }
 
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+   pgagroal_event_loop_break();
    return;
 
 client_error:
@@ -363,8 +357,8 @@ client_error:
    errno = 0;
 
    exit_code = WORKER_CLIENT_FAILURE;
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+
+   pgagroal_event_loop_break();
    return;
 
 server_error:
@@ -375,55 +369,44 @@ server_error:
    errno = 0;
 
    exit_code = WORKER_SERVER_FAILURE;
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+
+   pgagroal_event_loop_break();
    return;
 
 failover:
 
    exit_code = WORKER_FAILOVER;
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+
+   pgagroal_event_loop_break();
    return;
 
 get_error:
    pgagroal_log_warn("Failure during obtaining connection");
 
    exit_code = WORKER_SERVER_FAILURE;
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+
+   pgagroal_event_loop_break();
    return;
 }
 
 static void
-transaction_server(struct ev_loop* loop, struct ev_io* watcher, int revents)
+transaction_server(struct io_watcher* watcher)
 {
    int status = MESSAGE_STATUS_ERROR;
-   bool has_z = false;
    struct worker_io* wi = NULL;
    struct message* msg = NULL;
-   struct configuration* config = NULL;
+   struct main_configuration* config = NULL;
 
    wi = (struct worker_io*)watcher;
-   config = (struct configuration*)shmem;
-
-   /* We can't use the information from wi except from client_fd/client_ssl */
-   wi->server_fd = config->connections[slot].fd;
-   wi->slot = slot;
+   config = (struct main_configuration*)shmem;
 
    if (!pgagroal_socket_isvalid(wi->client_fd))
    {
       goto client_error;
    }
 
-   if (wi->server_ssl == NULL)
-   {
-      status = pgagroal_read_socket_message(wi->server_fd, &msg);
-   }
-   else
-   {
-      status = pgagroal_read_ssl_message(wi->server_ssl, &msg);
-   }
+   status = pgagroal_recv_message(watcher, &msg);
+
    if (likely(status == MESSAGE_STATUS_OK))
    {
       pgagroal_prometheus_network_received_add(msg->length);
@@ -442,8 +425,6 @@ transaction_server(struct ev_loop* loop, struct ev_io* watcher, int revents)
             {
                char tx_state = pgagroal_read_byte(msg->data + offset + 5);
 
-               has_z = true;
-
                if (tx_state != 'I' && !in_tx)
                {
                   pgagroal_prometheus_tx_count_add();
@@ -471,14 +452,8 @@ transaction_server(struct ev_loop* loop, struct ev_io* watcher, int revents)
          }
       }
 
-      if (wi->client_ssl == NULL)
-      {
-         status = pgagroal_write_socket_message(wi->client_fd, msg);
-      }
-      else
-      {
-         status = pgagroal_write_ssl_message(wi->client_ssl, msg);
-      }
+      status = pgagroal_send_message(watcher, msg);
+
       if (unlikely(status != MESSAGE_STATUS_OK))
       {
          goto client_error;
@@ -492,12 +467,18 @@ transaction_server(struct ev_loop* loop, struct ev_io* watcher, int revents)
          }
       }
 
-      if (!fatal)
+      /* Check for ReadyForQuery message (Z) to detect transaction completion */
+      if (msg->kind == 'Z' && !in_tx && slot != -1)
       {
-         if (has_z && !in_tx && slot != -1)
+         /* Transaction completed - stop I/O watcher immediately if still active */
+         if (io_watcher_active)
          {
-            ev_io_stop(loop, (struct ev_io*)&server_io);
+            pgagroal_io_stop(&server_io.io);
+            io_watcher_active = false;
+         }
 
+         if (!fatal)
+         {
             if (deallocate)
             {
                pgagroal_write_deallocate_all(wi->server_ssl, wi->server_fd);
@@ -512,15 +493,10 @@ transaction_server(struct ev_loop* loop, struct ev_io* watcher, int revents)
 
             slot = -1;
          }
-      }
-      else
-      {
-         if (has_z && !in_tx && slot != -1)
+         else
          {
-            ev_io_stop(loop, (struct ev_io*)&server_io);
-
             exit_code = WORKER_SERVER_FATAL;
-            running = 0;
+            pgagroal_event_loop_break();
          }
       }
    }
@@ -533,7 +509,6 @@ transaction_server(struct ev_loop* loop, struct ev_io* watcher, int revents)
       goto server_error;
    }
 
-   ev_break(loop, EVBREAK_ONE);
    return;
 
 client_error:
@@ -544,8 +519,8 @@ client_error:
    errno = 0;
 
    exit_code = WORKER_CLIENT_FAILURE;
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+
+   pgagroal_event_loop_break();
    return;
 
 server_done:
@@ -554,8 +529,7 @@ server_done:
                       strerror(errno), wi->server_fd, status);
    errno = 0;
 
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+   pgagroal_event_loop_break();
    return;
 
 server_error:
@@ -566,97 +540,108 @@ server_error:
    errno = 0;
 
    exit_code = WORKER_SERVER_FAILURE;
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+
+   pgagroal_event_loop_break();
    return;
 
 return_error:
    pgagroal_log_warn("Failure during connection return");
 
    exit_code = WORKER_SERVER_FAILURE;
-   running = 0;
-   ev_break(loop, EVBREAK_ALL);
+
+   pgagroal_event_loop_break();
    return;
 }
 
 static void
-start_mgt(struct ev_loop* loop)
+start_mgt(struct event_loop* loop __attribute__((unused)))
 {
-   memset(&io_mgt, 0, sizeof(struct ev_io));
-   ev_io_init(&io_mgt, accept_cb, unix_socket, EV_READ);
-   ev_io_start(loop, &io_mgt);
+   memset(&io_mgt, 0, sizeof(struct io_watcher));
+   pgagroal_event_accept_init(&io_mgt, unix_socket, accept_cb);
+   pgagroal_io_start(&io_mgt);
 }
 
 static void
-shutdown_mgt(struct ev_loop* loop)
+shutdown_mgt(struct event_loop* loop __attribute__((unused)))
 {
    char p[MISC_LENGTH];
-   struct configuration* config = NULL;
+   struct main_configuration* config = NULL;
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    memset(&p, 0, sizeof(p));
-   snprintf(&p[0], sizeof(p), ".s.%d", getpid());
+   snprintf(&p[0], sizeof(p), ".s.pgagroal.%d", getpid());
 
-   ev_io_stop(loop, &io_mgt);
+   pgagroal_io_stop(&io_mgt);
    pgagroal_disconnect(unix_socket);
    errno = 0;
    pgagroal_remove_unix_socket(config->unix_socket_dir, &p[0]);
    errno = 0;
+
+   if (server_io.io.msg)
+   {
+      free(server_io.io.msg->data);
+      free(server_io.io.msg);
+      server_io.io.msg = NULL;
+   }
 }
 
 static void
-accept_cb(struct ev_loop* loop, struct ev_io* watcher, int revents)
+accept_cb(struct io_watcher* watcher)
 {
-   struct sockaddr_in client_addr;
-   socklen_t client_addr_length;
-   int client_fd;
-   signed char id;
-   int32_t payload_slot;
-   int payload_i;
-   char* payload_s = NULL;
-   struct configuration* config = NULL;
+   int client_fd = -1;
+   int id = -1;
+   int32_t slot = -1;
+   int fd = -1;
+   struct main_configuration* config = NULL;
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
-   if (EV_ERROR & revents)
-   {
-      pgagroal_log_debug("accept_cb: invalid event: %s", strerror(errno));
-      errno = 0;
-      return;
-   }
-
-   client_addr_length = sizeof(client_addr);
-   client_fd = accept(watcher->fd, (struct sockaddr*)&client_addr, &client_addr_length);
+   client_fd = watcher->fds.main.client_fd;
    if (client_fd == -1)
    {
-      pgagroal_log_debug("accept: %s (%d)", strerror(errno), watcher->fd);
+      pgagroal_log_debug("accept: %s (%d)", strerror(errno), client_fd);
       errno = 0;
       return;
    }
 
    /* Process management request */
-   pgagroal_management_read_header(client_fd, &id, &payload_slot);
-   pgagroal_management_read_payload(client_fd, id, &payload_i, &payload_s);
-
-   switch (id)
+   if (pgagroal_connection_id_read(client_fd, &id))
    {
-      case MANAGEMENT_CLIENT_FD:
-         pgagroal_log_debug("pgagroal: Management client file descriptor: Slot %d FD %d", payload_slot, payload_i);
-         fds[payload_slot] = payload_i;
-         break;
-      case MANAGEMENT_REMOVE_FD:
-         pgagroal_log_debug("pgagroal: Management remove file descriptor: Slot %d FD %d", payload_slot, payload_i);
-         if (fds[payload_slot] == payload_i && !config->connections[payload_slot].new && config->connections[payload_slot].fd > 0)
-         {
-            pgagroal_disconnect(payload_i);
-            fds[payload_slot] = 0;
-         }
-         break;
-      default:
-         pgagroal_log_debug("pgagroal: Unsupported management id: %d", id);
-         break;
+      pgagroal_log_error("pgagroal: Management client: ID: %d", id);
+      goto done;
    }
 
+   if (id == CONNECTION_CLIENT_FD)
+   {
+      if (pgagroal_connection_transfer_read(client_fd, &slot, &fd))
+      {
+         pgagroal_log_error("pgagroal: Management client_fd: ID: %d Slot %d FD %d", id, slot, fd);
+         goto done;
+      }
+
+      fds[slot] = fd;
+   }
+   else if (id == CONNECTION_REMOVE_FD)
+   {
+      if (pgagroal_connection_transfer_read(client_fd, &slot, &fd))
+      {
+         pgagroal_log_error("pgagroal: Management remove_fd: ID: %d Slot %d FD %d", id, slot, fd);
+         goto done;
+      }
+
+      if (fds[slot] == fd && !config->connections[slot].new && config->connections[slot].fd > 0)
+      {
+         pgagroal_disconnect(fd);
+         fds[slot] = 0;
+      }
+   }
+   else
+   {
+      pgagroal_log_debug("pgagroal: Unsupported management id: %d", id);
+   }
+
+done:
+
    pgagroal_disconnect(client_fd);
 }
