@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2024 The pgagroal community
+ * Copyright (C) 2026 The pgagroal community
  *
  * Redistribution and use in source and binary forms, with or without modification,
  * are permitted provided that the following conditions are met:
@@ -28,6 +28,7 @@
 
 /* pgagroal */
 #include <pgagroal.h>
+#include <connection.h>
 #include <logging.h>
 #include <network.h>
 #include <management.h>
@@ -43,6 +44,7 @@
 
 /* system */
 #include <assert.h>
+#include <errno.h>
 #include <signal.h>
 #include <stdatomic.h>
 #include <stdbool.h>
@@ -57,6 +59,10 @@ static int find_best_rule(char* username, char* database);
 static bool remove_connection(char* username, char* database);
 static void connection_details(int slot);
 static bool do_prefill(char* username, char* database, int size);
+static bool is_alias_of_limit(char* database, int limit_index);
+static int get_connection_count_for_limit_rule(int rule_index, char* username);
+static char* resolve_database_name(char* database, int best_rule);
+static void check_graceful_shutdown_trigger(void);
 
 int
 pgagroal_get_connection(char* username, char* database, bool reuse, bool transaction_mode, int* slot, SSL** ssl)
@@ -72,12 +78,13 @@ pgagroal_get_connection(char* username, char* database, bool reuse, bool transac
    int best_rule;
    int retries;
    int ret;
+   char* real_database;
 
-   struct configuration* config;
-   struct prometheus* prometheus;
+   struct main_configuration* config;
+   struct main_prometheus* prometheus;
 
-   config = (struct configuration*)shmem;
-   prometheus = (struct prometheus*)prometheus_shmem;
+   config = (struct main_configuration*)shmem;
+   prometheus = (struct main_prometheus*)prometheus_shmem;
 
    pgagroal_prometheus_connection_get();
 
@@ -93,15 +100,6 @@ start:
    do_init = false;
    has_lock = false;
 
-   if (best_rule >= 0)
-   {
-      connections = atomic_fetch_add(&config->limits[best_rule].active_connections, 1);
-      if (connections >= config->limits[best_rule].max_size)
-      {
-         goto retry;
-      }
-   }
-
    connections = atomic_fetch_add(&config->active_connections, 1);
    has_lock = true;
    if (connections >= config->max_connections)
@@ -109,7 +107,6 @@ start:
       goto retry;
    }
 
-   /* Try and find an existing free connection */
    if (reuse)
    {
       for (int i = 0; *slot == -1 && i < config->max_connections; i++)
@@ -118,9 +115,22 @@ start:
 
          if (atomic_compare_exchange_strong(&config->states[i], &free, STATE_IN_USE))
          {
+            bool can_reuse = false;
+
+            // Check if same rule and username
             if (best_rule == config->connections[i].limit_rule &&
-                !strcmp((const char*)(&config->connections[i].username), username) &&
-                !strcmp((const char*)(&config->connections[i].database), database))
+                !strcmp((const char*)(&config->connections[i].username), username))
+            {
+               real_database = resolve_database_name(database, best_rule);
+               // Check exact database match
+               if (!strcmp((const char*)(&config->connections[i].database), real_database))
+               {
+                  can_reuse = true;
+               }
+               // Check if both are aliases of the same real database
+            }
+
+            if (can_reuse)
             {
                *slot = i;
             }
@@ -134,6 +144,15 @@ start:
 
    if (*slot == -1 && !transaction_mode)
    {
+      if (best_rule >= 0)
+      {
+         int rule_count = get_connection_count_for_limit_rule(best_rule, username);
+         if (rule_count >= config->limits[best_rule].max_size)
+         {
+            goto retry;
+         }
+      }
+
       /* Ok, try and create a new connection */
       for (int i = 0; *slot == -1 && i < config->max_connections; i++)
       {
@@ -164,6 +183,7 @@ start:
             if (!fork())
             {
                pgagroal_flush(FLUSH_GRACEFULLY, "*");
+               exit(0);
             }
 
             goto error;
@@ -181,7 +201,7 @@ start:
          }
          else
          {
-            ret = pgagroal_connect(config->servers[server].host, config->servers[server].port, &fd);
+            ret = pgagroal_connect(config->servers[server].host, config->servers[server].port, &fd, config->keep_alive, config->nodelay);
          }
 
          if (ret)
@@ -214,10 +234,11 @@ start:
 
          memset(&config->connections[*slot].username, 0, MAX_USERNAME_LENGTH);
          memcpy(&config->connections[*slot].username, username, MIN(strlen(username), MAX_USERNAME_LENGTH - 1));
-
+         real_database = resolve_database_name(database, best_rule);
          memset(&config->connections[*slot].database, 0, MAX_DATABASE_LENGTH);
-         memcpy(&config->connections[*slot].database, database, MIN(strlen(database), MAX_DATABASE_LENGTH - 1));
+         memcpy(&config->connections[*slot].database, real_database, MIN(strlen(real_database), MAX_DATABASE_LENGTH - 1));
 
+         pgagroal_log_debug("Connection setup: client_db='%s' -> postgres_db='%s'", database, real_database);
          config->connections[*slot].has_security = SECURITY_INVALID;
          config->connections[*slot].fd = fd;
 
@@ -274,7 +295,10 @@ start:
 
       config->connections[*slot].timestamp = time(NULL);
 
-      atomic_store(&prometheus->client_wait_time, difftime(time(NULL), start_time));
+      if (config->common.metrics > 0)
+      {
+         atomic_store(&prometheus->client_wait_time, difftime(time(NULL), start_time));
+      }
       pgagroal_prometheus_connection_success();
       pgagroal_tracking_event_slot(TRACKER_GET_CONNECTION_SUCCESS, *slot);
       pgagroal_prometheus_connection_unawaiting(best_rule);
@@ -339,12 +363,14 @@ retry2:
          {
             SLEEP_AND_GOTO(1000L, start)
          }
-
       }
    }
 
 timeout:
-   atomic_store(&prometheus->client_wait_time, difftime(time(NULL), start_time));
+   if (config->common.metrics > 0)
+   {
+      atomic_store(&prometheus->client_wait_time, difftime(time(NULL), start_time));
+   }
    pgagroal_prometheus_connection_timeout();
    pgagroal_tracking_event_basic(TRACKER_GET_CONNECTION_TIMEOUT, username, database);
    pgagroal_prometheus_connection_unawaiting(best_rule);
@@ -356,7 +382,10 @@ error:
       atomic_fetch_sub(&config->limits[best_rule].active_connections, 1);
    }
    atomic_fetch_sub(&config->active_connections, 1);
-   atomic_store(&prometheus->client_wait_time, difftime(time(NULL), start_time));
+   if (config->common.metrics > 0)
+   {
+      atomic_store(&prometheus->client_wait_time, difftime(time(NULL), start_time));
+   }
    pgagroal_prometheus_connection_error();
    pgagroal_prometheus_connection_unawaiting(best_rule);
    pgagroal_tracking_event_basic(TRACKER_GET_CONNECTION_ERROR, username, database);
@@ -368,12 +397,13 @@ int
 pgagroal_return_connection(int slot, SSL* ssl, bool transaction_mode)
 {
    int state;
-   struct configuration* config;
+   struct main_configuration* config;
    time_t now;
    signed char in_use;
    signed char age_check;
+   int transfer_fd = -1;
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    /* Kill the connection, if it lives longer than max_connection_age */
    if (config->max_connection_age > 0)
@@ -384,12 +414,12 @@ pgagroal_return_connection(int slot, SSL* ssl, bool transaction_mode)
       if (atomic_compare_exchange_strong(&config->states[slot], &in_use, age_check))
       {
          double age = difftime(now, config->connections[slot].start_time);
-         if ((age >= (double) config->max_connection_age && !config->connections[slot].tx_mode) ||
+         if ((age >= (double)config->max_connection_age && !config->connections[slot].tx_mode) ||
              !atomic_compare_exchange_strong(&config->states[slot], &age_check, STATE_IN_USE))
          {
             pgagroal_prometheus_connection_max_connection_age();
             pgagroal_tracking_event_slot(TRACKER_MAX_CONNECTION_AGE, slot);
-            return pgagroal_kill_connection(slot, ssl);
+            goto kill_connection;
          }
       }
    }
@@ -429,10 +459,42 @@ pgagroal_return_connection(int slot, SSL* ssl, bool transaction_mode)
 
          if (config->connections[slot].new)
          {
-            pgagroal_management_transfer_connection(slot);
+            if (pgagroal_connection_get(&transfer_fd))
+            {
+               goto kill_connection;
+            }
+
+            if (pgagroal_connection_id_write(transfer_fd, CONNECTION_TRANSFER))
+            {
+               goto kill_connection;
+            }
+
+            if (pgagroal_connection_transfer_write(transfer_fd, slot))
+            {
+               goto kill_connection;
+            }
+
+            pgagroal_disconnect(transfer_fd);
+            transfer_fd = -1;
          }
 
-         pgagroal_management_return_connection(slot);
+         if (pgagroal_connection_get(&transfer_fd))
+         {
+            goto kill_connection;
+         }
+
+         if (pgagroal_connection_id_write(transfer_fd, CONNECTION_RETURN))
+         {
+            goto kill_connection;
+         }
+
+         if (pgagroal_connection_slot_write(transfer_fd, slot))
+         {
+            goto kill_connection;
+         }
+
+         pgagroal_disconnect(transfer_fd);
+         transfer_fd = -1;
 
          if (config->connections[slot].limit_rule >= 0)
          {
@@ -446,6 +508,12 @@ pgagroal_return_connection(int slot, SSL* ssl, bool transaction_mode)
          atomic_store(&config->states[slot], STATE_FREE);
          atomic_fetch_sub(&config->active_connections, 1);
 
+         pgagroal_log_debug("Connection returned: slot=%d, active_connections=%d, gracefully=%s",
+                            slot, atomic_load(&config->active_connections), config->gracefully ? "true" : "false");
+
+         // Check for graceful shutdown after successful connection return
+         check_graceful_shutdown_trigger();
+
          pgagroal_prometheus_connection_return();
 
          return 0;
@@ -453,11 +521,16 @@ pgagroal_return_connection(int slot, SSL* ssl, bool transaction_mode)
       else if (state == STATE_GRACEFULLY)
       {
          pgagroal_write_terminate(ssl, config->connections[slot].fd);
+
+         // Check for graceful shutdown after graceful connection termination
+         check_graceful_shutdown_trigger();
       }
    }
 
 kill_connection:
 
+   pgagroal_disconnect(transfer_fd);
+
    pgagroal_tracking_event_slot(TRACKER_RETURN_CONNECTION_KILL, slot);
 
    return pgagroal_kill_connection(slot, ssl);
@@ -470,9 +543,10 @@ pgagroal_kill_connection(int slot, SSL* ssl)
    int ssl_shutdown;
    int result = 0;
    int fd;
-   struct configuration* config;
+   int transfer_fd;
+   struct main_configuration* config;
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    pgagroal_log_debug("pgagroal_kill_connection: Slot %d FD %d State %d PID %d",
                       slot, config->connections[slot].fd, atomic_load(&config->states[slot]),
@@ -483,7 +557,27 @@ pgagroal_kill_connection(int slot, SSL* ssl)
    fd = config->connections[slot].fd;
    if (fd != -1)
    {
-      pgagroal_management_kill_connection(slot, fd);
+      if (pgagroal_connection_get(&transfer_fd))
+      {
+         result = 1;
+      }
+
+      if (pgagroal_connection_id_write(transfer_fd, CONNECTION_KILL))
+      {
+         result = 1;
+      }
+
+      if (pgagroal_connection_slot_write(transfer_fd, slot))
+      {
+         result = 1;
+      }
+
+      if (pgagroal_connection_socket_write(transfer_fd, fd))
+      {
+         result = 1;
+      }
+
+      pgagroal_disconnect(transfer_fd);
 
       if (ssl != NULL)
       {
@@ -499,6 +593,10 @@ pgagroal_kill_connection(int slot, SSL* ssl)
 
       if (!pgagroal_socket_has_error(fd))
       {
+         if (pgagroal_socket_isvalid(config->connections[slot].fd))
+         {
+            pgagroal_write_terminate(NULL, config->connections[slot].fd);
+         }
          pgagroal_disconnect(fd);
       }
    }
@@ -515,6 +613,9 @@ pgagroal_kill_connection(int slot, SSL* ssl)
       }
 
       atomic_fetch_sub(&config->active_connections, 1);
+
+      // Check for graceful shutdown after killing connection
+      check_graceful_shutdown_trigger();
    }
 
    memset(&config->connections[slot].username, 0, sizeof(config->connections[slot].username));
@@ -555,12 +656,12 @@ pgagroal_idle_timeout(void)
    time_t now;
    signed char free;
    signed char idle_check;
-   struct configuration* config;
+   struct main_configuration* config;
 
    pgagroal_start_logging();
    pgagroal_memory_init();
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
    now = time(NULL);
    prefill = false;
 
@@ -614,12 +715,12 @@ pgagroal_max_connection_age(void)
    time_t now;
    signed char free;
    signed char age_check;
-   struct configuration* config;
+   struct main_configuration* config;
 
    pgagroal_start_logging();
    pgagroal_memory_init();
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
    now = time(NULL);
    prefill = false;
 
@@ -673,12 +774,12 @@ pgagroal_validation(void)
    time_t now;
    signed char free;
    signed char validation;
-   struct configuration* config;
+   struct main_configuration* config;
 
    pgagroal_start_logging();
    pgagroal_memory_init();
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
    now = time(NULL);
 
    pgagroal_log_debug("pgagroal_validation");
@@ -766,12 +867,12 @@ pgagroal_flush(int mode, char* database)
    signed char in_use;
    bool do_kill;
    signed char server_state;
-   struct configuration* config;
+   struct main_configuration* config;
 
    pgagroal_start_logging();
    pgagroal_memory_init();
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    prefill = false;
 
@@ -873,20 +974,18 @@ pgagroal_flush(int mode, char* database)
    pgagroal_pool_status();
    pgagroal_memory_destroy();
    pgagroal_stop_logging();
-
-   exit(0);
 }
 
 void
 pgagroal_flush_server(signed char server)
 {
-   struct configuration* config;
+   struct main_configuration* config;
    int primary = -1;
 
    pgagroal_start_logging();
    pgagroal_memory_init();
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    pgagroal_log_debug("pgagroal_flush_server %s", config->servers[server].name);
    for (int i = 0; i < config->max_connections; i++)
@@ -945,15 +1044,48 @@ pgagroal_flush_server(signed char server)
    exit(0);
 }
 
+void
+pgagroal_request_flush(SSL* ssl __attribute__((unused)), int client_fd, uint8_t compression, uint8_t encryption, struct json* payload)
+{
+   time_t start_time;
+   time_t end_time;
+   struct json* req = NULL;
+   int mode = 0;
+   char* database = NULL;
+
+   start_time = time(NULL);
+
+   req = (struct json*)pgagroal_json_get(payload, MANAGEMENT_CATEGORY_REQUEST);
+   mode = (int)pgagroal_json_get(req, MANAGEMENT_ARGUMENT_MODE);
+   database = (char*)pgagroal_json_get(req, MANAGEMENT_ARGUMENT_DATABASE);
+
+   if (database != NULL)
+   {
+      pgagroal_flush(mode, database);
+   }
+   else
+   {
+      pgagroal_flush(mode, "*");
+   }
+
+   end_time = time(NULL);
+
+   pgagroal_management_response_ok(NULL, client_fd, start_time, end_time, compression, encryption, payload);
+
+   pgagroal_json_destroy(payload);
+
+   exit(0);
+}
+
 void
 pgagroal_prefill(bool initial)
 {
-   struct configuration* config;
+   struct main_configuration* config;
 
    pgagroal_start_logging();
    pgagroal_memory_init();
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    pgagroal_log_debug("pgagroal_prefill");
 
@@ -1058,9 +1190,9 @@ pgagroal_prefill(bool initial)
 int
 pgagroal_pool_init(void)
 {
-   struct configuration* config;
+   struct main_configuration* config;
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    /* States */
    for (int i = 0; i < MAX_NUMBER_OF_CONNECTIONS; i++)
@@ -1088,9 +1220,9 @@ pgagroal_pool_init(void)
 int
 pgagroal_pool_shutdown(void)
 {
-   struct configuration* config;
+   struct main_configuration* config;
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    for (int i = 0; i < config->max_connections; i++)
    {
@@ -1122,9 +1254,9 @@ pgagroal_pool_shutdown(void)
 int
 pgagroal_pool_status(void)
 {
-   struct configuration* config;
+   struct main_configuration* config;
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    pgagroal_log_debug("pgagroal_pool_status: %d/%d", atomic_load(&config->active_connections), config->max_connections);
 
@@ -1144,18 +1276,42 @@ static int
 find_best_rule(char* username, char* database)
 {
    int best_rule;
-   struct configuration* config;
+   struct main_configuration* config;
 
    best_rule = -1;
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    if (config->number_of_limits > 0)
    {
       for (int i = 0; i < config->number_of_limits; i++)
       {
+         bool database_match = false;
+
+         // Check exact database name match or "all"
+         if (!strcmp("all", config->limits[i].database) || !strcmp(database, config->limits[i].database))
+         {
+            database_match = true;
+         }
+         else
+         {
+            // Check if database matches any alias for this limit entry
+            for (int j = 0; j < config->limits[i].aliases_count; j++)
+            {
+               if (!strcmp(database, config->limits[i].aliases[j]))
+               {
+                  database_match = true;
+                  pgagroal_log_debug("Database '%s' matched alias '%s' for rule %d",
+                                     database, config->limits[i].aliases[j], i);
+                  break;
+               }
+            }
+         }
+
+         // Check username match
+         bool username_match = (!strcmp("all", config->limits[i].username) || !strcmp(username, config->limits[i].username));
+
          /* There is a match */
-         if ((!strcmp("all", config->limits[i].username) || !strcmp(username, config->limits[i].username)) &&
-             (!strcmp("all", config->limits[i].database) || !strcmp(database, config->limits[i].database)))
+         if (username_match && database_match)
          {
             if (best_rule == -1)
             {
@@ -1164,7 +1320,8 @@ find_best_rule(char* username, char* database)
             else
             {
                if (!strcmp(username, config->limits[best_rule].username) &&
-                   !strcmp(database, config->limits[best_rule].database))
+                   (!strcmp(database, config->limits[best_rule].database) ||
+                    is_alias_of_limit(database, best_rule)))
                {
                   /* We have a precise rule already */
                }
@@ -1189,17 +1346,86 @@ find_best_rule(char* username, char* database)
       }
    }
 
+   if (best_rule != -1)
+   {
+      pgagroal_log_debug("find_best_rule: user=%s, database=%s -> rule %d (real_database=%s)",
+                         username, database, best_rule, config->limits[best_rule].database);
+   }
+
    return best_rule;
 }
 
+static bool
+is_alias_of_limit(char* database, int limit_index)
+{
+   struct main_configuration* config = (struct main_configuration*)shmem;
+
+   if (limit_index < 0 || limit_index >= config->number_of_limits)
+   {
+      return false;
+   }
+
+   for (int i = 0; i < config->limits[limit_index].aliases_count; i++)
+   {
+      if (!strcmp(database, config->limits[limit_index].aliases[i]))
+      {
+         return true;
+      }
+   }
+
+   return false;
+}
+
+static int
+get_connection_count_for_limit_rule(int rule_index, char* username)
+{
+   struct main_configuration* config = (struct main_configuration*)shmem;
+   int count = 0;
+
+   if (rule_index < 0 || rule_index >= config->number_of_limits)
+   {
+      return 0;
+   }
+
+   for (int i = 0; i < config->max_connections; i++)
+   {
+      if (atomic_load(&config->states[i]) != STATE_NOTINIT &&
+          !strcmp((const char*)(&config->connections[i].username), username))
+      {
+         // Check if this connection is for the main database name
+         if (!strcmp((const char*)(&config->connections[i].database), config->limits[rule_index].database))
+         {
+            count++;
+         }
+         else
+         {
+            // Check if this connection is for any alias of this database
+            for (int j = 0; j < config->limits[rule_index].aliases_count; j++)
+            {
+               if (!strcmp((const char*)(&config->connections[i].database), config->limits[rule_index].aliases[j]))
+               {
+                  count++;
+                  break;
+               }
+            }
+         }
+      }
+   }
+
+   pgagroal_log_debug("Connection count for rule %d (user=%s): %d connections",
+                      rule_index, username, count);
+
+   return count;
+}
+
 static bool
 remove_connection(char* username, char* database)
 {
    signed char free;
    signed char remove;
-   struct configuration* config;
+   struct main_configuration* config;
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    pgagroal_log_trace("remove_connection");
    for (int i = config->max_connections - 1; i >= 0; i--)
@@ -1232,16 +1458,75 @@ remove_connection(char* username, char* database)
    return false;
 }
 
+static void
+check_graceful_shutdown_trigger(void)
+{
+   struct main_configuration* config = (struct main_configuration*)shmem;
+
+   int remaining_connections = atomic_load(&config->active_connections);
+
+   if (remaining_connections == 0 && config->gracefully)
+   {
+      if (config->keep_running)
+      {
+         pgagroal_log_debug("pgagroal: graceful shutdown conditions met - connections=0, gracefully=true, keep_running=true - proceeding with shutdown");
+      }
+      else
+      {
+         pgagroal_log_debug("pgagroal: graceful shutdown skipped - connections=0, gracefully=true, but keep_running=false (race condition prevented)");
+         return;
+      }
+   }
+
+   if (remaining_connections == 0 && config->gracefully && config->keep_running)
+   {
+      // Double-check to handle race conditions with new connections
+      usleep(1000); // 1ms delay to let any concurrent operations complete
+
+      // Verify all three conditions are still met after brief delay
+      int final_connections = atomic_load(&config->active_connections);
+      if (final_connections == 0 && config->gracefully && config->keep_running)
+      {
+         pgagroal_log_debug("pgagroal: graceful shutdown confirmed - triggering shutdown");
+
+         pid_t parent_pid = getppid();
+         if (kill(parent_pid, SIGTRAP) == -1)
+         {
+            if (errno == ESRCH)
+            {
+               pgagroal_log_debug("Parent process already exited, worker shutting down");
+            }
+            else
+            {
+               pgagroal_log_debug("Failed to signal parent process (PID: %d): %s", parent_pid, strerror(errno));
+            }
+         }
+         else
+         {
+            pgagroal_log_debug("Graceful shutdown signal sent to parent process (PID: %d)", parent_pid);
+         }
+      }
+      else
+      {
+         int final_connections_check = atomic_load(&config->active_connections);
+         pgagroal_log_debug("Graceful shutdown conditions - connections=%d, gracefully=%s, keep_running=%s",
+                            final_connections_check,
+                            config->gracefully ? "true" : "false",
+                            config->keep_running ? "true" : "false");
+      }
+   }
+}
+
 static void
 connection_details(int slot)
 {
    int state;
    char time_buf[32];
    char start_buf[32];
-   struct configuration* config;
+   struct main_configuration* config;
    struct connection connection;
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
    connection = config->connections[slot];
    state = atomic_load(&config->states[slot]);
@@ -1446,28 +1731,43 @@ do_prefill(char* username, char* database, int size)
    signed char state;
    int free = 0;
    int connections = 0;
-   struct configuration* config;
+   struct main_configuration* config;
 
-   config = (struct configuration*)shmem;
+   config = (struct main_configuration*)shmem;
 
-   for (int i = 0; i < config->max_connections; i++)
+   //Find the limit rule for this database/username combination
+   int rule_index = find_best_rule(username, database);
+   if (rule_index != -1)
    {
-      if (!strcmp((const char*)(&config->connections[i].username), username) &&
-          !strcmp((const char*)(&config->connections[i].database), database))
+      // Count connections for this rule (including aliases)
+      connections = get_connection_count_for_limit_rule(rule_index, username);
+   }
+   else
+   {
+      // Fallback to old logic if no rule found
+      for (int i = 0; i < config->max_connections; i++)
       {
-         connections++;
-      }
-      else
-      {
-         state = atomic_load(&config->states[i]);
-
-         if (state == STATE_NOTINIT)
+         if (!strcmp((const char*)(&config->connections[i].username), username) &&
+             !strcmp((const char*)(&config->connections[i].database), database))
          {
-            free++;
+            connections++;
          }
       }
    }
 
+   // Count free connections
+   for (int i = 0; i < config->max_connections; i++)
+   {
+      state = atomic_load(&config->states[i]);
+      if (state == STATE_NOTINIT)
+      {
+         free++;
+      }
+   }
+
+   pgagroal_log_debug("do_prefill: user=%s, database=%s, current=%d, target=%d, free=%d",
+                      username, database, connections, size, free);
+
    return connections < size && free > 0;
 }
 
@@ -1497,3 +1797,26 @@ pgagroal_prefill_if_can(bool do_fork, bool initial)
       }
    }
 }
+
+static char*
+resolve_database_name(char* database, int best_rule)
+{
+   struct main_configuration* config = (struct main_configuration*)shmem;
+
+   if (best_rule < 0 || best_rule >= config->number_of_limits)
+   {
+      return database; // Return original if no rule found
+   }
+
+   // Check if this database name is an alias
+   for (int j = 0; j < config->limits[best_rule].aliases_count; j++)
+   {
+      if (!strcmp(database, config->limits[best_rule].aliases[j]))
+      {
+         return config->limits[best_rule].database; // Return real database name
+      }
+   }
+
+   // Not an alias, return original name
+   return database;
+}
\ No newline at end of file
